import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.awt.Color;
import java.awt.Point;
import java.io.*;
import java.util.*;

import javax.swing.JOptionPane;

/**
 * Write a description of class Graphs here.
 * 
 * @author Taylor Born
 * @version May 2012
 */
public class Graphs extends World
{
    public static Graphs main = new Graphs();
    private Menu menu;
    private Node selected = null;
    private boolean drag = false;
    
    private Point lastMouse = new Point(0, 0);
    
    private Connection highlight = null;
    
    private Connection editC;
    private InputBox inputB;
    
    private SPWindow spWindow;
    
    private static final int CODE_ID = 37;
    private ArrayList<Node> movedAround = new ArrayList<Node>();
    
    private enum Mode
    {
        CONNECT, DELETE, MOVE, WEIGHT
    }
    private Mode mode = Mode.CONNECT;

    public Graphs()
    {
        super(400, 300, 1);
        setPaintOrder(Menu.class, TextBox.class, Button.class, SPWindow.class, InputBox.class, Node.class, Label.class, Connection.class);
        main = this;
        setUpMenu();
        load();
    }
    private void setUpMenu()
    {
        ArrayList<String> menuStr = new ArrayList<String>(){{ add("Window"); add("Edit"); add("Tools"); }};
        menu = new Menu(menuStr, new Point(0, 0));
        menu.addItemsAsSet("Window/", new ArrayList<String>(){{ add("Calculator"); add("Y="); add("Graphs"); }}, "Graphs");
        menu.addItems("Edit/", new ArrayList<String>(){{ add("Color"); add("Reset Coloring"); add("Find Shortest Path"); add("Find Min. Spanning Tree"); add("Save"); add("Load"); add("Delete All"); }});
        menu.addItemsAsSet("Tools/", new ArrayList<String>(){{ add("Add and Connect"); add("Delete"); add("Move Vertices"); add("Set Path Weights"); }}, "Add and Connect");
        menu.addShortCut("Tools/Add and Connect", "a");
        menu.addShortCut("Tools/Delete", "d");
        menu.addShortCut("Tools/Move Vertices", "m");
        menu.addShortCut("Tools/Set Path Weights", "w");
        addObject(menu, 0, 0);
    }
    public void act()
    {
        MouseInfo mouse = Greenfoot.getMouseInfo();
        if (Greenfoot.mouseMoved(null) || Greenfoot.mouseDragged(null))
            lastMouse = new Point(mouse.getX(), mouse.getY());
        
        listenToMenu();
        
        if (editC != null)
        {
            if (inputB.isDone())
            {
                editC.setWeight(Integer.parseInt("0" + inputB.getInput()));
                inputB.remove();
                inputB = null;
                editC = null;
            }
            return;
        }
        
        if (spWindow != null && spWindow.getWorld() != null)
            return;
            
        List<Object> nodes = getObjects(Node.class);
        for (Object o : nodes)
        {
            Node oN = (Node)o;
            if (Greenfoot.mousePressed(oN))
            {
                selected = oN;
                drag = true;
            }
        }
        
        if (highlight != null)
        {
            highlight.remove();
            highlight = null;
        }
        if (!Greenfoot.mouseClicked(menu) && !Greenfoot.mousePressed(menu) && !menu.isActive())
        {
            switch (mode)
            {
                case CONNECT:   if (selected == null && Greenfoot.mousePressed(null))
                                {
                                    selected = new Node();
                                    addObject(selected, mouse.getX(), mouse.getY());
                                    removeColors();
                                    drag = true;
                                }
                                if (selected != null && Greenfoot.mouseDragEnded(null) || Greenfoot.mouseClicked(null))
                                {
                                    if (Math.sqrt(Math.pow(selected.getX() - mouse.getX(), 2) + Math.pow(selected.getY() - mouse.getY(), 2)) > 11)
                                    {
                                        boolean found = false;
                                        for (Object o : nodes)
                                        {
                                            Node oN = (Node)o;
                                            if (Greenfoot.mouseClicked(oN))
                                            {
                                                if (!selected.connectsTo(oN))
                                                {
                                                    addObject(new Connection(selected, oN), (selected.getX() + oN.getX()) / 2, (selected.getY() + oN.getY()) / 2);
                                                    removeColors();
                                                }
                                                found = true;
                                                break;
                                            }
                                        }
                                        if (!found)
                                        {
                                            Node nd = new Node();
                                            addObject(nd, mouse.getX(), mouse.getY());
                                            addObject(new Connection(selected, nd), (selected.getX() + nd.getX()) / 2, (selected.getY() + nd.getY()) / 2);
                                            removeColors();
                                        }
                                    }
                                    selected = null;
                                    drag = false;
                                }
                                break;
                case DELETE:    if (selected != null)
                                {
                                    if (Greenfoot.mousePressed(null))
                                    {
                                        selected.delete();
                                        selected = null;
                                        removeColors();
                                    }
                                }
                                else
                                    for (Object o : getObjects(Connection.class))
                                    {
                                        Connection c = (Connection)o;
                                        if (c.hits(lastMouse))
                                        {
                                            highlight = new Connection(c);
                                            addObject(highlight, c.getX(), c.getY());
                                            if (Greenfoot.mousePressed(null))
                                            {
                                                c.breakConnection();
                                                removeColors();
                                            }
                                            break;
                                        }
                                    }
                                break;
                case MOVE:      if (selected != null && Greenfoot.mouseDragged(null))
                                    selected.setLocation(mouse.getX(), mouse.getY());
                                if (Greenfoot.mouseClicked(null) || Greenfoot.mouseDragEnded(null))
                                {
                                    
                                    if (movedAround.size() < 3 && !movedAround.contains(selected))
                                    {
                                        movedAround.add(selected);
                                        if (movedAround.size() == 3)
                                            if (CodeGenerator.haveUser())
                                            {
                                                CodeDisplay cd = new CodeDisplay(CODE_ID);
                                                addObject(cd, cd.getImage().getWidth() / 2, getHeight() - 6);
                                            }
                                    }
                                    selected = null;
                                }
                                break;
                case WEIGHT:    for (Object o : getObjects(Connection.class))
                                {
                                    Connection c = (Connection)o;
                                    if (c.hits(lastMouse))
                                    {
                                        highlight = new Connection(c);
                                        addObject(highlight, c.getX(), c.getY());
                                        if (Greenfoot.mousePressed(null))
                                        {
                                            editC = c;
                                            inputB = new InputBox("Enter Weight:", c.getWeight() + "", "0123456789", 5);
                                            int y = (int)lastMouse.getY() - inputB.getImage().getHeight();
                                            if (lastMouse.getY() < getHeight() / 2)
                                                y = (int)lastMouse.getY() + inputB.getImage().getHeight();
                                            int x = (int)lastMouse.getX();
                                            if (x - inputB.getImage().getWidth() / 2 < 5)
                                                x = 5 + inputB.getImage().getWidth() / 2;
                                            else if (x + inputB.getImage().getWidth() / 2 > getWidth() - 5)
                                                x = getWidth() - 5 - inputB.getImage().getWidth() / 2;
                                            addObject(inputB, x, y);
                                        }
                                        break;
                                    }
                                }
                                break;
            }
        }
        else
        {
            selected = null;
            drag = false;
        }
        
    }
    private void listenToMenu()
    {
        String s = menu.getItemPressed();
        if (s != null)
        {
            if (s.startsWith("Window/"))
            {
                s = s.substring(s.indexOf("/") + 1);
                if (s.equals("Calculator/"))
                {
                    Screen.main.showCalculator();
                    Greenfoot.setWorld(Screen.main);
                }
                else if (s.equals("Y=/"))
                {
                    Screen.main.showXY();
                    Greenfoot.setWorld(Screen.main);
                }
                menu.provokeItem("Window/Graphs");
            }
            else if (s.startsWith("Edit/"))
            {
                s = s.substring(s.indexOf("/") + 1);
                if (s.equals("Color/"))
                    color();
                else if (s.equals("Reset Coloring/"))
                    removeColors();
                else if (s.equals("Find Shortest Path/"))
                    findShortest();
                else if (s.equals("Find Min. Spanning Tree/"))
                    findMinSpanningTree();
                else if (s.equals("Save/"))
                    save();
                else if (s.equals("Load/"))
                    load();
                else if (s.equals("Delete All/"))
                    deleteAll();
            }
            else if (s.startsWith("Tools/"))
            {
                s = s.substring(s.indexOf("/") + 1);
                if (s.equals("Add and Connect/"))
                    mode = Mode.CONNECT;
                else if (s.equals("Delete/"))
                    mode = Mode.DELETE;
                else if (s.equals("Move Vertices/"))
                    mode = Mode.MOVE;
                else if (s.equals("Set Path Weights/"))
                    mode = Mode.WEIGHT;
            }
        }
    }
    
    private Color[] colors = { Color.BLUE, Color.RED, Color.GREEN, Color.ORANGE, Color.CYAN, Color.MAGENTA, Color.YELLOW, Color.PINK, Color.BLACK };
    private void color()
    {
        List<Object> nodes = getObjects(Node.class);
        if (nodes.size() == 0)
            return;
        boolean swapped = true;
        while (swapped)
        {
            swapped = false;
            for (int i = 0; i < nodes.size() - 1; i++)
                if (((Node)nodes.get(i)).getConnectionCount() > ((Node)nodes.get(i + 1)).getConnectionCount())
                {
                    Object o = nodes.get(i);
                    nodes.set(i, nodes.get(i + 1));
                    nodes.set(i + 1, o);
                    swapped = true;
                }
        }
            
        ArrayList<Node> bank = new ArrayList<Node>();
        
        for (Object o : nodes)
        {
            Node oN = (Node)o;
            if (!oN.hasColor())
                color(oN);
            Node.ConnectionIterator ci = oN.getIterator();
            while (ci.hasNext())
            {
                Node n = ci.getNext();
                if (!n.hasColor() && !bank.contains(n))
                {
                    boolean inserted = false;
                    for (int i = 0; i < bank.size(); i++)
                        if (nodes.indexOf(bank.get(i)) > nodes.indexOf(n))
                        {
                            bank.add(i, n);
                            inserted = true;
                            break;
                        }
                    if (!inserted)
                        bank.add(n);
                }
            }
            while (!bank.isEmpty())
            {
                int ii = 0;
                int adjs = 0;
                for (int i = 0; i < bank.size(); i++)
                {
                    int a = 0;
                    ci = bank.get(i).getIterator();
                    while (ci.hasNext())
                    {
                        Node n = ci.getNext();
                        if (n.hasColor())
                            a++;
                    }
                    if (a > adjs)
                    {
                        ii = i;
                        adjs = a;
                    }
                }
                ci = bank.get(ii).getIterator();
                //ci = bank.get(0).getIterator();
                while (ci.hasNext())
                {
                    Node n = ci.getNext();
                    if (!n.hasColor() && !bank.contains(n))
                    {
                        boolean inserted = false;
                        for (int i = 0; i < bank.size(); i++)
                            if (nodes.indexOf(bank.get(i)) > nodes.indexOf(n))
                            {
                                bank.add(i, n);
                                inserted = true;
                                break;
                            }
                        if (!inserted)
                            bank.add(n);
                        //bank.add(n);
                    }
                }
                color(bank.remove(ii));
                //color(bank.remove(0));
            }
        }
    }
    private void color(Node n)
    {
        String s = "";
        Node.ConnectionIterator ci = n.getIterator();
        while (ci.hasNext())
        {
            Node nd = ci.getNext();
            if (nd.hasColor())
                s += " " + nd.getColor() + " ";
        }
        int c = 0;
        while (s.contains(" " + c + " "))
            c++;
        n.color(c, colors[c]);
    }
    private void removeColors()
    {
        List<Object> nodes = getObjects(Node.class);
        for (Object o : nodes)
            ((Node)o).resetColor();
    }
    private void findMinSpanningTree()
    {
        List<Object> connections = getObjects(Connection.class);
        boolean swapped = true;
        while (swapped)
        {
            swapped = false;
            for (int i = 0; i < connections.size() - 1; i++)
            {
                if (((Connection)connections.get(i)).getWeight() > ((Connection)connections.get(i + 1)).getWeight())
                {
                    Object o = connections.get(i);
                    connections.set(i, connections.get(i + 1));
                    connections.set(i + 1, o);
                    swapped = true;
                }
            }
        }
        ArrayList<ArrayList<Node>> nodes = new ArrayList<ArrayList<Node>>();
        for (int i = 0; i < connections.size(); i++)
        {
            Connection c = (Connection)connections.get(i);
            boolean found = false;
            for (int k = 0; k < nodes.size(); k++)
            {
                if (nodes.get(k).contains(c.getN1()) || nodes.get(k).contains(c.getN2()))
                    found = true;
                if (!nodes.get(k).contains(c.getN1()) || !nodes.get(k).contains(c.getN2()))
                {
                    if (nodes.get(k).contains(c.getN1()))
                    {
                        boolean found2 = false;
                        for (int kk = 0; kk < nodes.size(); kk++)
                            if (kk != k && nodes.get(kk).contains(c.getN2()))
                            {
                                for (Node n : nodes.get(kk))
                                    nodes.get(k).add(n);
                                nodes.remove(kk);
                                if (k > kk)
                                    k--;
                                found = true;
                                break;
                            }
                        if (!found2)
                            nodes.get(k).add(c.getN2());
                        connections.remove(i--);
                    }
                    else if (nodes.get(k).contains(c.getN2()))
                    {
                        boolean found2 = false;
                        for (int kk = 0; kk < nodes.size(); kk++)
                            if (kk != k && nodes.get(kk).contains(c.getN1()))
                            {
                                for (Node n : nodes.get(kk))
                                    nodes.get(k).add(n);
                                nodes.remove(kk);
                                if (k > kk)
                                    k--;
                                found = true;
                                break;
                            }
                        if (!found2)
                            nodes.get(k).add(c.getN1());
                        connections.remove(i--);
                    }
                }
            }
            if (!found)
            {
                nodes.add(new ArrayList<Node>());
                nodes.get(nodes.size() - 1).add(c.getN1());
                nodes.get(nodes.size() - 1).add(c.getN2());
                connections.remove(i--);
            }
        }
        for (Object o : connections)
            ((Connection)o).breakConnection();
        removeColors();
    }
    private void findShortest()
    {
        if (editC == null && (spWindow == null || spWindow.getWorld() == null))
        {
            spWindow = new SPWindow();
            addObject(spWindow, getWidth() / 2, getHeight() / 2);
        }
    }
    public String findShortestPath(Node nStart, Node nEnd)
    {
        if (nStart == nEnd)
            return nStart.getLabelName() + "\nTotalWeight: 0";
        List<Object> nodes = getObjects(Node.class);
        ArrayList<Node> setNodes = new ArrayList<Node>();
        Integer[] nodeLabels = new Integer[nodes.size()];
        Node[] comeFroms = new Node[nodes.size()];
        
        nodeLabels[nodes.indexOf(nStart)] = 0;
        Node n = nStart;
        
        while (n != nEnd)
        {
            Node.ConnectionIterator ci = n.getIterator();
            while (ci.hasNext())
            {
                int w = ci.getThisWeight();
                int i = nodes.indexOf(ci.getNext());
                int la = nodeLabels[nodes.indexOf(n)] + w;
                if (nodeLabels[i] == null || la < nodeLabels[i])
                {
                    nodeLabels[i] = la;
                    comeFroms[i] = n;
                }
            }
            setNodes.add(n);
            int min = -1;
            int minI = -1;
            for (int i = 0; i < nodeLabels.length; i++)
                if (nodeLabels[i] != null && !setNodes.contains(nodes.get(i)) && (minI == -1 || nodeLabels[i] < min))
                {
                    min = nodeLabels[i];
                    minI = i;
                }
            if (minI == -1)
                break;
            n = (Node)nodes.get(minI);
        }
        
        Integer i = nodeLabels[nodes.indexOf(nEnd)];
        if (i != null)
        {
            String s = " to " + nEnd.getLabelName() + "\nTotal Weight: " + i;
            n = comeFroms[nodes.indexOf(nEnd)];
            while (n != nStart)
            {
                s = " to " + n.getLabelName() + s;
                n = comeFroms[nodes.indexOf(n)];
            }
            s = nStart.getLabelName() + s;
            return s;
        }
        return "No Path exists.";
    }
    public Node getNode(String s)
    {
        List<Object> nodes = getObjects(Node.class);
        for (Object o : nodes)
            if (((Node)o).getLabelName().equals(s))
                return (Node)o;
        return null;
    }
    private void load()
    {
        deleteAll();
        try
        {
            BufferedReader br = new BufferedReader(new InputStreamReader(getClass().getClassLoader().getResource("Graph.txt").openStream()));
            String str = br.readLine();
            
            ArrayList<Node> nodes = new ArrayList<Node>();
            while (!str.isEmpty())
            {
                int w = 1;
                if (str.length() > 1 && str.charAt(1) >= '0' && str.charAt(1) <= '9')
                    w = 2;
                nodes.add(new Node(str.substring(0, w)));
                str = str.substring(w);
            }
            
            double d = 360d / nodes.size();
            double dd = 180;
            for (int i = 0; i < nodes.size(); i++)
            {
                addObject(nodes.get(i), getWidth() / 2 + (int)(Math.cos(Math.toRadians(dd)) * 130), getHeight() / 2 + (int)(Math.sin(Math.toRadians(dd)) * 130));
                dd -= d;
            }
            String edges = br.readLine();
            while (!edges.isEmpty())
            {
                int w = 1;
                if (edges.charAt(1) >= '0' && edges.charAt(1) <= '9')
                    w = 2;
                Node n1 = getNode(edges.substring(0, w));
                edges = edges.substring(w);
                w = 1;
                if (edges.charAt(1) >= '0' && edges.charAt(1) <= '9')
                    w = 2;
                Node n2 = getNode(edges.substring(0, w));
                addObject(new Connection(n1, n2), (n1.getX() + n2.getX()) / 2, (n1.getY() + n2.getY()) / 2);
                edges = edges.substring(edges.indexOf(",") + 1);
            }
            String weights = br.readLine();
            List<Connection> connections = getObjects(Connection.class);
            int i = 0;
            while (!weights.isEmpty())
            {
                ((Connection)connections.get(i++)).setWeight(Integer.parseInt(weights.substring(0, weights.indexOf(","))));
                
                weights = weights.substring(weights.indexOf(",") + 1);
            }
        }
        catch (IOException e)
        {}
    }
    public void save()
    {
        List<Object> nodes = getObjects(Node.class);
        for (Object o : nodes)
            System.out.print(((Node)o).getLabelName());
        System.out.println("");
        List<Object> connections = getObjects(Connection.class);
        String weights = "";
        for (Object o : connections)
        {
            Connection c = (Connection)o;
            System.out.print(c.getN1().getLabelName() + c.getN2().getLabelName() + ",");
            weights += c.getWeight() + ",";
        }
        System.out.println("\n" + weights);
    }
    private void deleteAll()
    {
        removeObjects(getObjects(Node.class));
        removeObjects(getObjects(Connection.class));
        removeObjects(getObjects(Label.class));
        Node.resetCount();
    }
}